import os
import uuid
from mimetypes import guess_type

from abejacli.session import (
    generate_user_session,
    generate_retry_session,
)
from abejacli.config import (
    PLATFORM_BASE_URL,
    FILE_READ_CHUNK_SIZE,
    PLATFORM_REQUEST_TIMEOUT_SECONDS,
)
from abejacli.datalake.process_file_job import (
    INITIALIZE_REPORT,
    PROGRESS_REPORT,
    FINISH_REPORT,
    RAISE_ERROR
)


class UploadFileIterator(object):

    def __init__(self, file_id, path, publisher_id, report_queue, chunk_size=FILE_READ_CHUNK_SIZE):
        self.file_id = file_id
        self.path = path
        self.publisher_id = publisher_id
        self.report_queue = report_queue
        self.chunk_size = chunk_size
        self.total_size = os.path.getsize(path)

    def __iter__(self):
        initialize_options = {
            'file_name': os.path.basename(self.path),
            'total': self.total_size,
        }
        self.report_queue.put((INITIALIZE_REPORT, self.publisher_id, 0, initialize_options))
        with open(self.path, 'rb') as upload_file:
            while True:
                data = upload_file.read(self.chunk_size)
                if not data:
                    break
                # update progress with chunk data size
                self.report_queue.put((PROGRESS_REPORT, self.publisher_id, len(data), None))
                yield data
        result_options = {
            'source': self.path,
            'destination': self.file_id,
        }
        self.report_queue.put((FINISH_REPORT, self.publisher_id, 0, result_options))

    def __len__(self):
        return self.total_size


class IterableToFileAdapter(object):

    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)
        self.buffer = b''
        self.total_read = 0

    def read(self, size):
        result = self.buffer
        while len(result) < size:
            try:
                retrieved = next(self.iterator)
            except StopIteration:
                break
            result = b''.join([result, retrieved])
        result, self.buffer = result[:size], result[size:]
        self.total_read += len(result)
        return result

    def __len__(self):
        return self.length


def upload_job(channel_id, file_path, report_queue, options):
    """
    upload files until consuming all items in file queue

    :param channel_id: channel identifier
    :param file_path: path to file
    :param report_queue: queue to report progress for each file
    :param options: job options
    :return:
    """

    publisher_id = uuid.uuid4().hex

    try:
        url = "{}/channels/{}".format(PLATFORM_BASE_URL, channel_id)
        type, _ = guess_type(file_path)
        headers = {
            'x-abeja-meta-filename': os.path.basename(file_path),
            'Content-Type': type if type else 'application/octet-stream'
        }
        metadata = options.get('metadata', ()) if options else ()
        metadata_header = {
            'x-abeja-meta-{}'.format(key): value for key, value in metadata
        }
        headers.update(metadata_header)
        # Get upload url
        with generate_user_session() as session:
            res = session.post(url, None, headers=headers, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        res.raise_for_status()
        res_json = res.json()
        upload_url = res_json['upload_url']
        file_id = res_json['file_id']
        # File iterator
        it = UploadFileIterator(file_id, file_path, publisher_id, report_queue)
        data_adapter = IterableToFileAdapter(it)
        with generate_retry_session() as upload_session:
            upload_res = upload_session.put(upload_url, data=data_adapter, headers=headers,
                                            timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        upload_res.raise_for_status()
    except:
        options = {
            'source': file_path,
            'error': 'Failed to upload {} to channel_id {}'.format(file_path, channel_id)
        }
        report_queue.put((RAISE_ERROR, publisher_id, 0, options))
